#!/usr/bin/env python
"""osctrtask.py 
   - Task to bias-subtract OSMOS setup and slit images, display them,
     and derive an offset for the guide probe to center the target.
   - This version is invoked as a task from within pyraf.
"""

import sys
from pyraf import iraf
from iraf import images, imstatistics, imdelete, imarith, imcopy, imtile, hselect
from iraf import noao, imred, bias, linebias, imexamine, imgets, hedit
from iraf import imcoords, ccfind
import os
from time import sleep, time
from cooclasses import *


guiderscale = 3.4821  # guider steps per unbinned pixel on the OSMOS chip.

def xformtoguider(xi, eta, rotator) :  # this may not be needed after all.

   eta = eta * MIS_Magic['GPARITY'] / MIS_Magic['GSCALE']
   xi = xi / MIS_Magic['GSCALE']

   rotangle = MIS_Magic['GROTATION'] + MIS_Magic['GROTANGSENSE'] * rotator
   cosrot = cos(rotangle / _skysub.DEG_IN_RADIAN)
   sinrot = sin(rotangle / _skysub.DEG_IN_RADIAN)

   tempxi  = xi *  cosrot  +  eta * sinrot
   tempeta = eta * cosrot  -  xi *  sinrot

   tempxi = tempxi + MIS_Magic['GXCENT']
   tempeta = tempeta + MIS_Magic['GYCENT']

   return [tempxi,tempeta]

#   Vestigial, from the standalone non-pyraf version.
#
# class osctrconfig :  # simple container for fixed information.
#   def __init__(self) :
#      try :
#         inf = open("osctr.config","r")
#      except :
#         print "Couldn't open 'osctr.config'.  Will use defaults."
#         self.datapath = "/data/hiltner/"
#         self.prefix = "ms1."
# #        self.slitx = 2060.
# #        self.slity = 1989.
#         return
#      for l in inf.readlines() :
#         x = l.split()
#         if len(x) > 0 :
#            if x[0] == "datapath" :  self.datapath = x[1]
#            if x[0] == "prefix" : self.prefix = x[1]
# #           if x[0] == "slitx" : self.slitx = float(x[1])
# #           if x[0] == "slity" : self.slity = float(x[1])
#      inf.close()
 
 
def ds9_is_open() :
   is_open = True
   tmptmp = os.popen("xpaaccess ds9")  # returns a file-like object of output
   for l in tmptmp.readlines() :
      if l.find("no") > -1 :
          is_open = False
   tmptmp.close()
   return is_open

def biasquads(imroot, outroot) :

   # Kill any scratch files dead - sometimes imdelete leaves them in limbo and
   # breaks the task.

   # clear with 'rm' ... imdelete doesn't always kill 'em.

   os.system("rm %s.fits" % outroot)
   os.system("rm scratch1.fits")
   os.system("rm scratch2.fits")
   os.system("rm scratch3.fits")
   os.system("rm scratch4.fits")
   
   # Split out each quadrant and linebias it.
   
   # First, figure out how to slice up image based on header info -- 
   # Have to assume that the point where the quadrants meet is dead-center.
   # Also, we'll assume that there's an overscanx on the outer side of
   # each image.

   hselout = hselect(imroot,"naxis1,naxis2,overscnx","yes",Stdout=1)
   x = hselout[0].split()
   naxis1 = int(x[0])
   naxis2 = int(x[1])
   overscanx = int(x[2])
   maxx1 = naxis1 / 2  # largest x for this quadrant
   minx2 = maxx1 + 1   # min x for next quadrant
   maxx2 = naxis1
   maxy1 = naxis2 / 2
   miny2 = maxy1 + 1
   maxy2 = naxis2

   # cut into scratch files - 
   # scratch 1 - lowx, lowy
   # scratch 2 - lowx, highy
   # scratch 3 - highx, lowy
   # scratch 4 - highx, highy

   imcopy(imroot + "[1:%d,1:%d]" % (maxx1,maxy1),"scratch1")   
   imcopy(imroot + "[1:%d,%d:%d]" % (maxx1,miny2,maxy2),"scratch2")
   imcopy(imroot + "[%d:%d,1:%d]" % (minx2,maxx2,maxy1),"scratch3")
   imcopy(imroot + "[%d:%d,%d:%d]" % (minx2,maxx2,miny2,maxy2),"scratch4")
   
   # linebias each sub-image:
   
   linebias.function = "leg"
   linebias.order = 1
   linebias.niterate = 2
   linebias.interactive = 0
   linebias.median = 0
   
   linebias.trim = "[*,*]"  # i.e., don't trim - retain exact geometry to be safe.
   # first and second images - bias strip at low x
   linebias.bias = "[3:%d,*]" % (overscanx - 3)   
   linebias("scratch1","scratch1")
   linebias("scratch2","scratch2")
   # third and fourth - bias is at high x
   linebias.bias = "[%d:%d,*]" % (maxx1 - 3, maxx1 - overscanx + 3)
   linebias("scratch3","scratch3")
   linebias("scratch4","scratch4")
   
   # write a list of the sub-images, for feeding into imtile -- this
   # ordering puts them together correctly, with the imtile params specified 
   # below.
   scratchlist = open("scratchlist","w")
   scratchlist.write("scratch1\n")
   scratchlist.write("scratch3\n")
   scratchlist.write("scratch2\n")
   scratchlist.write("scratch4\n")
 #  scratchlist.write("scratch1\n")
 #  scratchlist.write("scratch3\n")
 #  scratchlist.write("scratch4\n")
 #  scratchlist.write("scratch2\n")
   scratchlist.close()
   
 # specify imtile params explicitly, or tiling may get
 # scrambled or corrupted.
   imtile.nctile = 2
   imtile.nltile = 2
   imtile.trim_section= "[*,*]"
   imtile.start_tile = "ll"
   imtile.row_order = "yes"
   imtile.raster_order = "no"
   imtile.subtract = "no"
   imtile.ncols = "INDEF"
   imtile.nlines = "INDEF"
   imtile.ncoverlap = 0
   imtile.nloverlap = 0
   imtile.verbose = "yes"

   imtile("@scratchlist",outroot,2,2)

# copy the most relevant header commands.  Hedit doesn't seem to
# offer a nice way of doing this.

   hselout = hselect(imroot,"RA,dec,equinox,rotangle,filtid2,object,exptime,date-obs,time-obs,jd","yes",Stdout=1)
   rastr, decstr, eqstr, rotangstr, filtid2str, objectstr, exptimestr, datestr, timestr, jdstr = hselout[0].split('\t')
   hedit(outroot,'RA',rastr,add = 'yes', update = 'yes', verify = 'no', show = 'no')
   decstr = decstr.replace('+','')   # idiotic, won't take an explicit +.
   hedit(outroot,'dec',decstr.strip(),add = 'yes', update = 'yes', verify = 'no', show = 'no')
   hedit(outroot,'equinox',eqstr,add = 'yes', update = 'yes', verify = 'no', show = 'no')
   hedit(outroot,'object',objectstr,add = 'yes', update = 'yes', verify = 'no', show = 'no')
   hedit(outroot,'rotangle',rotangstr,add = 'yes', update = 'yes', verify = 'no', show = 'no')
   rotangle = float(rotangstr)
   hedit(outroot,'filtid2',filtid2str,add = 'yes', update = 'yes', verify = 'no', show = 'no')
   hedit(outroot,'exptime',exptimestr,add = 'yes', update = 'yes', verify = 'no', show = 'no')
   hedit(outroot,'date-obs',datestr,add = 'yes', update = 'yes', verify = 'no', show = 'no')
   hedit(outroot,'time-obs',timestr,add = 'yes', update = 'yes', verify = 'no', show = 'no')
   print "jdstr = ",jdstr   # make our own obs epoch to avoid that stupid asthedit kluge to keep JD digits ...
   jd = float(jdstr)
   epoch = 2000. + (jd - 2451545.0) / 365.25
   # write an 'obsep' keyword for PM correction, if implemented.
   hedit(outroot,"obsep",epoch,add='yes',update='yes',verify='no', show = 'no')
   hedit(outroot,"secpix1",0.273,add='yes',update='yes',verify='no', show = 'no')
   # keep a copy of the original source image name 
   hedit(outroot,"orig-im",imroot,add='yes',update='yes',verify='no', show = 'no')

   return (epoch, rotangle)  # a couple of things that are good to have in main prog

def parseimexam(imexamout) :
   """return the coordinates of the slit and the 
    star, from the output of imexam. """

   xslit = -1000.
   yslit = -1000.
   xstar = -1000.
   ystar = -1000.
   for l in imexamout :
       x = l.split()
       if len(x) == 3 :
          xslit = float(x[0])
          yslit = float(x[1])
       elif len(x) == 4 :
          try :
             xstar = float(x[0])
             ystar = float(x[1])
          except :
             pass
   return [xslit, yslit, xstar, ystar]

# Start of main MAIN Main program

def osctrtask(acqnum, slitnum, datapath, prefix, slitns, twoframes, 
   xslitpar, yslitpar, getwcs,
   marktarg, autotarg, fromlist, targetra, targetdec, 
   targlist, targname, rawtarg, pushprobe) :
    
   if not ds9_is_open() :
      os.system("ds9 &")
      sleep(2)  # give it time to open
#    
#    try :
#       acqnum = int(sys.argv[1])
#    except :
#       print __doc__
#       print "no image specified, dying."
#       sys.exit()
#    
#    twoframes = False
#    try :
#       slitnum = int(sys.argv[2])
#       twoframes = True
#    except :
#       print "Single frame specified."
#    
#   config = osctrconfig()

   acqroot = datapath + prefix + "%04d" % (acqnum)
   if twoframes :
       slitroot = datapath + prefix + "%04d" % (slitnum)
       print "slitroot = ",slitroot
   
   # account for possible binning
   xbinning = 1
   ybinning = 1
   try :
      binhsel = hselect(acqroot,"ccdxbin,ccdybin","yes",Stdout=1)
      # acquisition and slit images must be binned identically ... 
      x = binhsel[0].split()
      xbinning = int(x[0])
      ybinning = int(x[1])
      print "x, y binning: ",xbinning,ybinning
   except :
      print "CAUTION: Couldn't read ccdxbin, ccdybin values from %s." % \
            (acqroot)

   dobias = True   # this was a parameter, but you always want it.
        
   if dobias :
      # make the bias-subbed acquisition image 
      obsep, rotangle = biasquads(acqroot,"acqimage")
      print "through biasquads, obseq and rotangle = ",obsep, rotangle
      # grab the epoch and rotator angle while you're at it.
   else :
      os.system("rm acqimage.fits") # done insite biasquads otherwise
      imcopy(acqroot,"acqimage")
      
   
   autotargOK = False  # needs initialization
   if getwcs :   # Try for a wcs solution if asked. 
      # grab info to fake a .hsel file for acquisition image.
      
      hselout = hselect(acqroot,"ra,dec,equinox,naxis1,naxis2","yes",Stdout=1)
      hselfile = open("acqimage.hsel","w")   # apply calib to generic bias-subbed imaged ...
      hselfile.write("%s %5.3f\n" % (hselout[0],0.2730 * xbinning))
      hselfile.close()
      
      #obsepout = hselect(acqroot,"OBSEP","yes",Stdout=1)
      #print "obsepout = ",obsepout

      #rotangout = hselect(acqroot,"rotangle","yes",Stdout =1 )
      # print "rotangout = ",rotangout
      #rotangle = float(rotangout[0])
      
      # sex it ... 
      os.system("sex -c /usr/local/pkg/thorsrc/coordmatch/starfind.sex acqimage.fits\n");
      os.system("cat2seeing acqimage")
      # match it
      if slitns : orientstr = "EW"   # if slit N/S, long axis of acqimage is E/W.
      else : orientstr = "NS"   # if slit E/W, long axis of acqimage is N/S.
     
      os.system("matchusno_oblong acqimage --noquery --catalog ucac --rotator %6.1f --longaxis %s --epoch %8.2f\n" % 
                  (rotangle, orientstr, obsep));

      # check to see if it wrote a solution, based on time last modified.
      matchOK = False
      autotargOK = False
      time_ago = 1000.
      # bring time_ago out of loop so I can see error msg
      time_ago = time() - os.path.getmtime('acqimage.radec.match.wcs')
      try : 
          time_ago = time() - os.path.getmtime('acqimage.radec.match.wcs')
          if time_ago < 30. :  matchOK = True
          else : print "Coord solution is %f sec old, rejected!" % (time_ago)
      except :
          print "time_ago calculation failed; time_ago = %f" % (time_ago)
           # (leaves matchOK = False)
      

#      print "issued command: "
#      print "matchusno_oblong acqimage --noquery --catalog ucac --rotator %6.1f --longaxis %s --epoch %8.2f" % \
#                  (rotangle, orientstr, obsep);
      
#      print "Trying UCAC3 for better match ... "
#      os.system("usno2ucac3 acqimage %f" % (obsepoch))
      
      if matchOK : 
         try :
            os.system("./wcs_script1")
            print "WCS has been loaded into acqimage."
         except :
            print "WCS loading script didn't execute."
         
         sleep(2)  # to give wcs time to load ...
      else :
          print "Match apparently failed, so WCS not loaded to acqimage."
   
   # display the bias-subbed acquisition image
   os.system('xpaset -p ds9 frame frameno 1')  # force it to be in Frame 1
   os.system('xpaset -p ds9 file acqimage.fits')

   if getwcs and marktarg and matchOK:
      coordsok = False   # get coords from parameters as strings
      fileOK = False
      if fromlist : # get object coords from a list.
         try :
            listf = open(targlist, "r") 
            fileOK = True
         except :
            print "Couldn't open target list '%s'." % listf
         if fileOK :
            for l in listf.readlines() :
               if not coordsok :  # not yet found
                  xx = l.split()
                  if xx[0] == targname :  # exact match
                     try :
                         targetra = xx[1] + " " + xx[2] + " " + xx[3]
                         targetdec = xx[4] + " " + xx[5] + " " + xx[6]
                         # osctrtask.setParam("targetra",targetra)
                         targeteq = xx[7]
                         cel = celest([targetra, targetdec, targeteq])
                         print "Will look for target at %s, %s [%s]" % (targetra, targetdec, targeteq)
                         coordsok = True
                     except :
                         print "Coords for ",targname," in ",targlist," did not parse!"
                         print "line reads: ",l
      else :  # get object coords from display.
         try :  # interpret with cooclasses, which is more permissive than iraf.
            print "targetra, targetdec = ",targetra,targetdec
            cel = celest([targetra,targetdec,2000.])   # targetra and dec are strings.
            coordsok = True
         except :
            print "Couldn't interpret the target coords, so won't mark 'em."

      if coordsok :  
         # format strings for xpaset ra and dec input
         rastr = cel.ra.tripletstring(delin=":",showsign="no")
         decstr = cel.dec.tripletstring(delin=":",showsign="yes")
         print "ra, dec format strings: ",rastr,decstr
         # draw a circle on the target coordinates.  Syntax is hairy.
         os.system('echo "icrs; circle %s %s 0.15\'" | xpaset ds9 regions' % 
               (rastr,decstr)) 
         if autotarg : 
            # crazy rigamarole needed for this because ccfind does I/O through files ...
            xpixpred = -1000.
            ypixpred = -1000.
            try : os.remove('scratchradec')
            except OSError : pass
            try : os.remove('scratchxy')
            except OSError : pass
            f1 = open('scratchradec','w')
            f1.write("%s %s\n" % (rastr,decstr))
            f1.close()
            ccfind('scratchradec','scratchxy','acqimage',lngcolumn=1,latcolumn=2,usewcs="yes")
            f2 = open("scratchxy",'r')
            for l in f2.readlines() :
               xx = l.split()
               if len(xx) > 0 :
                  if xx[0].find("#") < 0 :
                     try :
                         xpixpred = float(xx[2])
                         ypixpred = float(xx[3])
                         autotargOK = True
                     except : 
                         pass
            f2.close()
            if autotargOK :
               # imexam also operates through a file.
               f3 = open("myimcurcmds","w")
               f3.write("%f %f 101\n" % (xpixpred,ypixpred))
               f3.write("q\n")
               f3.close()
               imexamout = imexamine("acqimage", imagecur = "myimcurcmds",use_display = "no",  
                   defkey = 'a', Stdout = 1)
               imexparsed = False
               print "Automatic run of imexam returned the following:"
               for ll in imexamout :
                   print "ll = ", ll
                   xx = ll.split()
                   if len(xx) == 4:  # the right line has length exactly 4.
                       try :
                          xstar = float(xx[0])
                          ystar = float(xx[1])
                          imexparsed = True
                       except :
                          pass
               if not imexparsed :
                   print "Automatic target centroiding didn't work ... "
                   autotargOK = False
        
   elif marktarg :
      # Explain, in case someone tries to mark without wcs.
      print "NOTE: Can't mark or automark the target, since WCS is not being derived."

   if twoframes :
      if dobias :
         biasquads(slitroot,"slitimage")
      else :
         os.system("rm slitimage.fits")
         imcopy(slitroot,"slitimage")
      os.system('xpaset -p ds9 frame frameno 2')
      os.system('xpaset -p ds9 file slitimage.fits')
      slitFound = False

   else :
      xslit = xslitpar
      yslit = yslitpar
      print "No slit image specified."
      print "Using slit position as supplied: %6.1f, %6.1f" % (xslit,yslit)
      slitFound = True

   if autotargOK : targOK = True
   else : targOK = False
   if rawtarg : targOK = False
   

   while not slitFound or not targOK :
      print "On DS9 display, "
      if not slitFound :
         print "Mark slit with 'x',", 
      if rawtarg :
         print " manually note target XY,"
      elif not targOK :
         print "Mark target with 'a',",
      print "'q' quits."
      print "(Use ds9 'frame' commands to switch images.)"
      imexamout = imexamine(Stdout = 1)
      [xslitout,yslitout,xstarout,ystarout] = parseimexam(imexamout)
      if not slitFound :
         if xslitout == -1000. or yslitout == -1000. : 
            print """Didn't get the slit coordinates.  Try again;
               Mark slit with an 'x'."""
         else  :  
            xslit = xslitout
            yslit = yslitout 
            slitFound = True

      if rawtarg :
         while not targOK :
            print "Type in target X and Y, space separated:"
            inline = raw_input()
            xx = inline.split()
            try :
               xstar = float(xx[0])
               ystar = float(xx[1])
               targOK = True
            except :
               print "Didn't parse, try again."

      if not targOK : 
         if xstarout == -1000. or ystarout == -1000. :
            print """Didnt' get the target coordinates.  Try again;
               Mark star with an 'a'. """
         else :
            xstar = xstarout
            ystar = ystarout
            targOK = True
   
   print "xslit, yslit = %4.1f, %4.1f" % (xslit,yslit)
   print "xstar, ystar = %4.1f, %4.1f" % (xstar,ystar)
   
   dxpix = xslit - xstar  
   dypix = yslit - ystar
   
   if slitns :   # TESTED 2010 Sept - signs OK.
      dxguider = dypix * guiderscale * xbinning
      dyguider = dxpix * guiderscale * ybinning
   else :        # know to work for slit EW.
      dxguider = dxpix * guiderscale * xbinning
      dyguider = -1. * dypix * guiderscale * ybinning
   
   print "Set guider DX: %4.0f  DY: %4.0f" % (dxguider, dyguider)
   
   # end of main routine ...
   
   print "... and move telescope to put guide star back in the box."

parfile = iraf.osfn("/lhome/obs24m/iraf/uparm/osctrtask.par")
t = iraf.IrafTaskFactory(taskname = "osctrtask", value = parfile,
     function = osctrtask)

